/*
 *  lol_types.cpp
 *
 *  LOL++, implementation of LOLCODE-like language on top of C++
 *  Copyright (C) 2008-2009 Yusdi Saliman
 *
 *  This file contains the implementation of the C++ class that emulates
 *  LOLCODE's data types.
 *
 *  LOL++ is free software. It comes without any warranty, to
 *  the extent permitted by applicable law. You can redistribute it
 *  and/or modify it under the terms of the Do What The Fuck You Want
 *  To Public License, Version 2, as published by Sam Hocevar. See
 *  http://sam.zoy.org/wtfpl/COPYING for more details.
 */

#define __LOL_TYPES_CPP
#include "lol_types.h"
#include <sstream>
using namespace std;

__LolData::__LolData(const char* str)
{
    type = __LOL_YARN;
    strVal = str;
    pArray = NULL;
    size = 0;
}

__LolData::__LolData(string str)
{
    type = __LOL_YARN;
    strVal = str;
    pArray = NULL;
    size = 0;
}

__LolData::__LolData(int n)
{
    type = __LOL_NUMBR;
    intVal = n;
    pArray = NULL;
    size = 0;
}

__LolData::__LolData(double f)
{
    type = __LOL_NUMBAR;
    floatVal = f;
    pArray = NULL;
    size = 0;
}

__LolData::__LolData(bool b)
{
    type = __LOL_TROOF;
    boolVal = b;
    pArray = NULL;
    size = 0;
}
    
bool __LolData::operator== (const __LolData& lol)
{
    if (this->type != lol.type)
        return false;
    
    switch (this->type)
    {
        case __LOL_NOOB:
            return true;
        case __LOL_YARN:
            return (this->strVal == lol.strVal);
        case __LOL_NUMBR:
            return (this->intVal == lol.intVal);
        case __LOL_NUMBAR:
            return (this->floatVal == lol.floatVal);
        case __LOL_TROOF:
            return (this->boolVal == lol.boolVal);
        default:
            return false;
    }
}

bool __LolData::operator==(const char* str)
{
    return (type == __LOL_YARN && strVal == str);
}

bool __LolData::operator==(int n)
{
    // it's possible to compare float with int
    if (type == __LOL_NUMBR)
        return (intVal == n);
    else if (type == __LOL_NUMBAR)
        return (floatVal == (double) n);
    else
        return false;
}

bool __LolData::operator==(double f)
{
    // it's possible to compare float with int
    if (type == __LOL_NUMBR)
        return (intVal == (int) f);
    else if (type == __LOL_NUMBAR)
        return (floatVal == f);
    else
        return false;
}

bool __LolData::operator==(bool b)
{
    return (type == __LOL_TROOF && boolVal == b);
}

bool __LolData::operator!=(const __LolData& lol)
{
    return !(*this == lol);
}

bool __LolData::operator!=(const char* str)
{
    return !(*this == str);
}

bool __LolData::operator!=(int n)
{
    return !(*this == n);
}

bool __LolData::operator!=(double f)
{
    return !(*this == f);
}

bool __LolData::operator!=(bool b)
{
    return !(*this == b);
}

// add operator&& to make BOTH SAEM works
__LolData __LolData::operator&&(__LolData lol)
{
    return this->toBool() && lol.toBool();
}

// string concatenation operator
__LolData __LolData::operator&(__LolData lol)
{
    string str = toString();
    str += lol.toString();
    __LolData ret(str);
    return ret;
}

/** check that the second operand is either integer or float, and do neccessary
  * changes in case the operation involves both integer and float. */
bool __LolData::checkTypes(__LolData& lol)
{
    // we only accept integer and float
    if ((type != __LOL_NUMBR && type != __LOL_NUMBAR) ||
        (lol.type != __LOL_NUMBR && lol.type != __LOL_NUMBAR))
        return false;
    
    // any operation with float will result in a float
    if (lol.type == __LOL_NUMBAR && type == __LOL_NUMBR)
    {
        type = __LOL_NUMBAR;
        floatVal = (double) intVal;
    }
    else if (lol.type == __LOL_NUMBR && type == __LOL_NUMBAR)
    {
        lol.type = __LOL_NUMBAR;
        lol.floatVal = (double) lol.intVal;
    }
    
    return true;
}

/** Add integer or float number to this object's value */
void __LolData::add(__LolData lol)
{
    if (!checkTypes(lol))
        return;

    switch (type)
    {
        case __LOL_NUMBR:
            intVal += lol.intVal;
        case __LOL_NUMBAR:
            floatVal += lol.floatVal;
        default:
            return;
    }
}

/** Substract integer or float number from this object's value */
void __LolData::sub(__LolData lol)
{
    if (!checkTypes(lol))
        return;

    switch (type)
    {
        case __LOL_NUMBR:
            intVal -= lol.intVal;
        case __LOL_NUMBAR:
            floatVal -= lol.floatVal;
        default:
            return;
    }
}

/** Multiply integer or float number with this object's value */
void __LolData::multiply(__LolData lol)
{
    if (!checkTypes(lol))
        return;

    switch (type)
    {
        case __LOL_NUMBR:
            intVal *= lol.intVal;
        case __LOL_NUMBAR:
            floatVal *= lol.floatVal;
        default:
            return;
    }
}

/** Divide integer or float number from this object's value */
void __LolData::divide(__LolData lol)
{
    if (!checkTypes(lol))
        return;

    switch (type)
    {
        case __LOL_NUMBR:
            intVal /= lol.intVal;
        case __LOL_NUMBAR:
            floatVal /= lol.floatVal;
        default:
            return;
    }
}

/** Modulus operation with this object's value */
void __LolData::mod(__LolData lol)
{
    // for modulus, the type can only be integer
    if (type != __LOL_NUMBR || lol.type != __LOL_NUMBR)
        return;

    intVal = intVal % lol.intVal;
}

/** Change to a new data type */
void __LolData::castTo(__LOL_TYPES t)
{
    switch (t)
    {
        case __LOL_YARN:
            strVal = toString();
            break;
        case __LOL_NUMBR:
            intVal = toInt();
            break;
        case __LOL_NUMBAR:
            floatVal = toFloat();
            break;
        case __LOL_TROOF:
            boolVal = toBool();
            break;
    }
    
    type = t;
}

/** return a string representation of the data */
string __LolData::toString()
{
    ostringstream stm;
    
    switch (type)
    {
        case __LOL_NOOB:
            return "NOOB";
        case __LOL_YARN:
            return strVal;
        case __LOL_NUMBR:
            stm << intVal;
            return stm.str();
        case __LOL_NUMBAR:
            stm << floatVal;
            return stm.str();
        case __LOL_TROOF:
            return boolVal ? "WIN" : "FAIL";
        default:
            return "";
    }
}

/** return an integer representation of the data */
int __LolData::toInt()
{
    istringstream stm(strVal);
    
    switch (type)
    {
        case __LOL_NOOB:
            return 0;
        case __LOL_YARN:
            stm >> intVal;
            // fall through
        case __LOL_NUMBR:
            return intVal;
        case __LOL_NUMBAR:
            return (int) floatVal;
        case __LOL_TROOF:
            return (int) boolVal;
        default:
            return 0;
    }
}

/** return a floating point number representation of the data */
double __LolData::toFloat()
{
    istringstream stm(strVal);
    
    switch (type)
    {
        case __LOL_NOOB:
            return 0.0;
        case __LOL_YARN:
            stm >> floatVal;
            // fall through
        case __LOL_NUMBAR:
            return floatVal;
        case __LOL_NUMBR:
            return (double) intVal;
        case __LOL_TROOF:
            return (double) boolVal;
        default:
            return 0.0;
    }
}

/** return a boolean representation of the data */
bool __LolData::toBool()
{
    istringstream stm(strVal);
    
    switch (type)
    {
        case __LOL_NOOB:
            return false;
        case __LOL_YARN:
            stm >> boolVal;
            // fall through
        case __LOL_TROOF:
            return boolVal;
        case __LOL_NUMBR:
            return (bool) intVal;
        case __LOL_NUMBAR:
            return (bool) floatVal;
        default:
            return false;
    }
}

/** get a reference to the data in a specified index in the array */
__LolData& __LolData::operator[] (__LolData index)
{
    if (!pArray)
        createArray(index);
    else if (size < index.toInt())
        resizeArray(index);

    return pArray[index.toInt()];
}

/** create array */
void __LolData::createArray(__LolData s)
{
    pArray = new __LolData[s.toInt() + 1];
    size = s.toInt();
}

/** resize array */
void __LolData::resizeArray(__LolData s)
{
    __LolData* pNewArray = new __LolData[s.toInt() + 1];
    int i;

    if (s.toInt() <= size)
    {
        for (i = 0; i <= s.toInt(); i++)
            pNewArray[i] = pArray[i];
    }
    else
    {
        for (i = 0; i <= size; i++)
            pNewArray[i] = pArray[i];
    }

    delete [] pArray;
    pArray = pNewArray;
    size = s.toInt();
}

/** Copy constructor */
__LolData::__LolData(const __LolData& lol)
{
    type = lol.type;
    strVal = lol.strVal;
    intVal = lol.intVal;
    floatVal = lol.floatVal;
    boolVal = lol.boolVal;

    if (lol.pArray)
    {
        size = lol.size;
        pArray = new __LolData[size + 1];
        for (int i = 0; i <= size; i++)
           pArray[i] = lol.pArray[i];
    }
    else
    {
        size = 0;
        pArray = NULL;
    }
}

//// assignment operator

__LolData& __LolData::operator= (__LolData lol)
{
    if (this == &lol)
        return *this;

    if (pArray)
        delete [] pArray;

    type = lol.getType();
    strVal = lol.toString();
    intVal = lol.toInt();
    floatVal = lol.toFloat();
    boolVal = lol.toBool();

    if (lol.getLength().toInt() > 0)
    {
        size = lol.getLength().toInt();
        pArray = new __LolData[size + 1];
        for (int i = 0; i <= size; i++)
           pArray[i] = lol[i];
    }
    else
    {
        size = 0;
        pArray = NULL;
    }

    return *this;
}

__LolData& __LolData::operator= (const char* str)
{
    type = __LOL_YARN;
    strVal = str;
    if (pArray)
        delete [] pArray;
    pArray = NULL;
    size = 0;
}

__LolData& __LolData::operator= (string str)
{
    type = __LOL_YARN;
    strVal = str;
    if (pArray)
        delete [] pArray;
    pArray = NULL;
    size = 0;
}

__LolData& __LolData::operator= (int n)
{
    type = __LOL_NUMBR;
    intVal = n;
    if (pArray)
        delete [] pArray;
    pArray = NULL;
    size = 0;
}

__LolData& __LolData::operator= (double f)
{
    type = __LOL_NUMBAR;
    floatVal = f;
    if (pArray)
        delete [] pArray;
    pArray = NULL;
    size = 0;
}

__LolData& __LolData::operator= (bool b)
{
    type = __LOL_TROOF;
    boolVal = b;
    if (pArray)
        delete [] pArray;
    pArray = NULL;
    size = 0;
}

